/* Smart curtain project */
/* (c) VCCH 2022 */
// includes
//#define DEBG
#include <Arduino.h>
#include <Adafruit_NeoPixel.h>
#include <WiFi.h>
#include <HTTPClient.h>
#include "image.h"

// Defines & parameters
#define REDLIGHTR 4 // Light reduction to save power
#define REDLIGHTG 4 // Light reduction to save power
#define REDLIGHTB 4 // Light reduction to save power
#define NB_TRIES 1 // Nb tries wifi before going forward
#define PIN 26
#define BPIN 39
#define NUMPIXELS (60*20)
#define NUMFLASH (NUMPIXELS/6)
#define NSPEED 30
#define FSPEED 5
#define SSPEED 60
#define UPDELAY ((1000*5))
#define NB_MODES 13
#define STAGES 10  // stage speed
#define MINSTAGES 2 // Min stage speed
#define WSIZE 15 // Worm size
#define PSIZE 4 // PAC size
#define PACNUM 22 // pacman blobs
#define COLRS 0
#define ADCS 10
#define DURCLOWN (3000/NSPEED)
#define DURSPRITE (100/NSPEED)
#define SIZEX 20
#define SIZEY 60
#define STARTMODE 12
#define OFFLED 8

// enter your wifi credentials here
const char* ssid = ** fill it in** ;
const char* password = ** fill it in**  ;
const char* staturl =*** fill it in ***"http://your domain .com/state.txt";

// Variables global
int dspeed=NSPEED;
int colr=0;
int stage=0;
int adc=ADCS;
int color=1;
int shade;
int count=0;
int nmode=1;
int rseed=0;
int state[NUMPIXELS];
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ400);
HTTPClient http;

// Routines
void gethttpars(){
    if ((WiFi.status() == WL_CONNECTED)) {
      http.begin(staturl);
      int httpCode = http.GET();
      if (httpCode > 0) {
        String payload = http.getString();
        #ifdef DEBG
        Serial.println(httpCode);
        Serial.println(payload);
        #endif
        int pmode=(payload[0]-'0');
        if (pmode>=NB_MODES){
          switch(pmode){
            case 13:
              dspeed=NSPEED;
            break;
            case 14:
              dspeed=FSPEED;
            break;}
          }
        else{
          nmode=(pmode!=color);
          color=pmode;}
      } else {
        #ifdef DEBG
        Serial.println("Error on HTTP request");
        #endif
        }
      http.end();}
}

void cycling(){
  colr+=adc;
  if ((colr>255)||(colr<0)){
    colr-=adc;
    adc=-adc;}
  int j=colr;
  int addb=adc;
  for (int i=0;i<NUMPIXELS;++i){
    pixels.setPixelColor(i, pixels.Color(j*(shade&1), j*((shade>>1)&1), j*((shade>>2)&1)));
    j+=addb;
    if ((j>255)||(j<0)){
    j-=addb;
    addb=-addb;}
    }
}

void dispmap(const uint8_t *ptr,bool sens){
      for (int numpix=0;numpix<NUMPIXELS;++numpix){
        int pixy=numpix%SIZEY; // 0=bottom  SIZEY-1=top
        int pixx=sens?numpix/SIZEY:SIZEX-1-numpix/SIZEY; // 0=right SIZEX-1=left
        
        int picx=pixx; // 0=left SIZEX-1=right  Invert if sens=false
        int picy=(picx&1)?pixy:(SIZEY-1)-pixy; // 0=top SIZEY-1=bottom
        int addr=3*(picx+picy*SIZEX); // adress in the C array : X increase left to right, Y increase top to bottom
        unsigned char cred=ptr[addr]/REDLIGHTR;
        unsigned char cgreen=ptr[addr+1]/REDLIGHTG;
        unsigned char cblue=ptr[addr+2]/REDLIGHTB;
        if (pixy<SIZEY-OFFLED)
          pixels.setPixelColor(numpix, pixels.Color(cred,cgreen,cblue));}
    pixels.show();
  }

void clowncycling(){
  const uint8_t *ptr;
  stage=stage+1;
  if ((stage%DURCLOWN)==0){
    shade+=1;
    switch(shade%2){
      case 0:
        ptr=palm1;
        break;
      case 1:
        ptr=mondrian1;
        break;}
   dispmap(ptr,true);
  }
}

// sprt1 : immobile face sprt2 : marche devant sprt3 : immobile profil sprt 4 : marche profil
void sprite(){
  const uint8_t *ptr;
  stage=stage+1;
  if ((stage%DURSPRITE)==0){
    shade+=1;
    bool sens=true;
    switch(shade%14){
      case 0:
      case 2:
      case 4:
      case 9:
        ptr=sprt1;
        break;
      case 1:
      case 3:
        ptr=sprt2;
        sens=(shade%4)==1;
        break;
      case 5:
      case 7:
        ptr=sprt3;
        break;
      case 6:
      case 8:
        ptr=sprt4;
        break;
      case 10:
      case 12:
        ptr=sprt3;
        sens=false;
        break;
      case 11:
      case 13:
        ptr=sprt4;
        sens=false;
        break;
 }
   dispmap(ptr,sens);
  }
}

void worm(){
  for (int i=0;i<NUMPIXELS;++i){
     int j=255-(255*(i-stage)/WSIZE);
     if ((j<=255)&&(j>=0))
       pixels.setPixelColor(i, pixels.Color(j,j,j));}      
  stage=stage-1;
  if (stage<0)
      stage=NUMPIXELS+WSIZE-1;
}

void paclean(){
  for (int i=0;i<NUMPIXELS;++i){
    if ((i%PACNUM)!=0){
      pixels.setPixelColor(i, pixels.Color(0, 0, 0));}
    else{
      pixels.setPixelColor(i, pixels.Color(255, 255, 255));}}
  pixels.show();
}  

void pworm(){
  for (int i=0;i<NUMPIXELS;++i){
     int j=255-(255*(i-stage)/(PSIZE+(stage&1)));
     if ((j<=255)&&(j>=-(255/PSIZE)))
       pixels.setPixelColor(i, pixels.Color(max(0,j),max(j,0),0));}      
  stage=stage-1;
  if (stage<0){
      paclean();
      stage=NUMPIXELS+WSIZE-1;}
}

void pclean(){
  for (int i=0;i<NUMPIXELS;++i)
    pixels.setPixelColor(i, pixels.Color(0, 0, 0));  
  pixels.show();
}


void flashing(){
  int nblit=0;
  // Update state
  for (int i=0;i<NUMPIXELS;++i){
    if ( (STAGES<state[i]) && (state[i]<(511-STAGES))){
      int j=state[i]>255?max(0,511-state[i]):state[i];
      pixels.setPixelColor(i, pixels.Color(j,j,j));
      state[i]+=random(MINSTAGES,STAGES);
      nblit++;}
    else{
      if (state[i]!=0){
        state[i]=0;
        pixels.setPixelColor(i, pixels.Color(0, 0, 0));}}
   }
   int tolight=NUMFLASH-nblit;
   if (tolight>0){
     tolight=random(0,2*tolight);
    for (;tolight>0;--tolight)
      state[random(0, NUMPIXELS-1)]=random(1+STAGES,500-STAGES);}
}

// Setup
void setup() {
  int i=0;
  for (i=0;i<NUMPIXELS;i++) state[i]=0;
#ifdef DEBG
Serial.begin(115200);
#endif
  pinMode(BPIN, INPUT);
  WiFi.begin(ssid, password);
  i=0;
  while ((i<NB_TRIES)&&(WiFi.status() != WL_CONNECTED)) {
    i+=1;
    delay(1000);
#ifdef DEBG
    Serial.println("Connecting to WiFi..");
#endif
  }
#ifdef DEBG
  if (WiFi.status() == WL_CONNECTED)
 Serial.println("Connected to the WiFi network");
#endif
 color=STARTMODE;
 pixels.begin();
 pixels.clear();
}

// main Loop
void loop() {
  // Color upload loop
  count+=dspeed;
  if (count>UPDELAY){
    #ifdef DEBG
     Serial.println(color);
    #endif
    count=0;
    gethttpars();
    }
    
  // button test
  if (digitalRead(BPIN) == LOW){
    color=(color+1)%NB_MODES;
    delay(300);
    #ifdef DEBG
    Serial.println(color);
    #endif
    
    nmode=1; // new mode instead some setup is needed
  }
  // Color display
  switch(color){
    case 0:
      if(nmode)
       pclean();
     break;
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
     if (nmode){
       shade=color;
       colr=COLRS;
       adc=ADCS;}
     cycling();
     pixels.show();
     break;
    case 8: // Worm
     if (nmode){
       stage=0;
       pclean();}
     worm();
     pixels.show();
     break;
    case 9: // Star
     if (nmode) 
      pclean();
     flashing();
     pixels.show();
     break;
    case 10: // Tutti
     if (nmode){
        pixels.clear();
       shade=0;
       stage=0;}
     sprite();
     break;
    case 11: // Pac
     if (nmode){
       stage=0;
       paclean();}
     pworm();
     pixels.show();
     break;
     case 12: // Clown
      if (nmode){
        pixels.clear();
        stage=0;
        shade=0;}
      clowncycling();
      break;
}
  nmode=0;
  delay(dspeed);
}